<?php

namespace WHMCS\Module\Registrar\Hosttech;

use WHMCS\Module\Registrar\Hosttech\Hosttech;

/**
 * PHP Class which loads the required registrars for the HP Domainchecker module.
 *
 * @copyright  2018 HOSTTECH GmbH
 */
class LoadRegistrars
{
    private $registrars;

    const REGISTRAR_MIN_VERSION = "1.0.53";
    /*
     *  Constructor
     */
    public function __construct()
    {
        $this->registrars = array();
        $this->loadHOSTTECHRegistrars();
    }

    /*
     * Returns all the loaded registrars
     *
     * @return array All loaded registrars
     */
    public function getLoadedRegistars()
    {
        return $this->registrars;
    }

    /*
     * Loads all the HOSTTECH Registrars.
     */
    private function loadHOSTTECHRegistrars()
    {
        foreach ($this->getAllConfiguredRegistrars() as $registrar) {
            $this->loadSingleHOSTTECHRegistrar($registrar);
        }
        //if no registrar configured in the pricelist, then try to add hosttech and hosttech
        if (empty($this->registrars)) {
            $this->loadSingleHOSTTECHRegistrar("hosttech");
            $this->loadSingleHOSTTECHRegistrar("hosttech");
        }
    }

    /*
     * Loads a single HOSTTECH Registrar
     * HOSTTECH registrar = HOSTTECH like registrar module >= 1.0.53 AND registrar module authentication successful

     * @param string $registrar The name of the registrar
     */
    private function loadSingleHOSTTECHRegistrar($registrar)
    {
        if (isset($registrar) && in_array($registrar, ["hosttech", "hosttech"])) {
            //compare version number
            $version = Hosttech::getRegistrarModuleVersion($registrar);
            if (version_compare($version, self::REGISTRAR_MIN_VERSION) >= 0) {
                //check authentication
                $checkAuthentication = Hosttech::call(["COMMAND" => "CheckAuthentication"]);
                if ($checkAuthentication["CODE"] == "200") {
                    array_push($this->registrars, $registrar);
                }
            }
        }
    }

    /*
     * Returns all the WHMCS registrars
     *
     * @return array List of all WHMCS registrars
     */
    private function getAllConfiguredRegistrars()
    {
        $list = array();
        $registrars_array = Helper::SQLCall("SELECT extension, autoreg FROM tbldomainpricing GROUP BY autoreg", array(), "fetchall");
        foreach ($registrars_array as $registrar) {
            array_push($list, $registrar["autoreg"]);
        }
        return $list;
    }
}
